import { computed, ref } from 'vue-demi';
import { resolveUnref, clamp, reactify } from '@vueuse/shared';

function createGenericProjection(fromDomain, toDomain, projector) {
  return (input) => {
    return computed(() => projector(resolveUnref(input), resolveUnref(fromDomain), resolveUnref(toDomain)));
  };
}

const defaultNumericProjector = (input, from, to) => {
  return (input - from[0]) / (from[1] - from[0]) * (to[1] - to[0]) + to[0];
};
function createProjection(fromDomain, toDomain, projector = defaultNumericProjector) {
  return createGenericProjection(fromDomain, toDomain, projector);
}

function logicAnd(...args) {
  return computed(() => args.every((i) => resolveUnref(i)));
}

function logicNot(v) {
  return computed(() => !resolveUnref(v));
}

function logicOr(...args) {
  return computed(() => args.some((i) => resolveUnref(i)));
}

function useAbs(value) {
  return computed(() => Math.abs(resolveUnref(value)));
}

function resolveUnrefArgsFlat(args) {
  return args.flatMap((i) => {
    const v = resolveUnref(i);
    if (Array.isArray(v))
      return v.map((i2) => resolveUnref(i2));
    return [v];
  });
}

function useAverage(...args) {
  return computed(() => {
    const array = resolveUnrefArgsFlat(args);
    return array.reduce((sum, v) => sum += v, 0) / array.length;
  });
}

function useCeil(value) {
  return computed(() => Math.ceil(resolveUnref(value)));
}

function useClamp(value, min, max) {
  const _value = ref(value);
  return computed({
    get() {
      return _value.value = clamp(_value.value, resolveUnref(min), resolveUnref(max));
    },
    set(value2) {
      _value.value = clamp(value2, resolveUnref(min), resolveUnref(max));
    }
  });
}

function useFloor(value) {
  return computed(() => Math.floor(resolveUnref(value)));
}

function useMath(key, ...args) {
  return reactify(Math[key])(...args);
}

function useMax(...args) {
  return computed(() => {
    const array = resolveUnrefArgsFlat(args);
    return Math.max(...array);
  });
}

function useMin(...args) {
  return computed(() => {
    const array = resolveUnrefArgsFlat(args);
    return Math.min(...array);
  });
}

function useProjection(input, fromDomain, toDomain, projector) {
  return createProjection(fromDomain, toDomain, projector)(input);
}

function useRound(value) {
  return computed(() => Math.round(resolveUnref(value)));
}

function useSum(...args) {
  return computed(() => resolveUnrefArgsFlat(args).reduce((sum, v) => sum += v, 0));
}

function useTrunc(value) {
  return computed(() => Math.trunc(resolveUnref(value)));
}

export { logicAnd as and, createGenericProjection, createProjection, logicAnd, logicNot, logicOr, logicNot as not, logicOr as or, useAbs, useAverage, useCeil, useClamp, useFloor, useMath, useMax, useMin, useProjection, useRound, useSum, useTrunc };
